"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _keys = require("babel-runtime/core-js/object/keys");

var _keys2 = _interopRequireDefault(_keys);

var _stringify = require("babel-runtime/core-js/json/stringify");

var _stringify2 = _interopRequireDefault(_stringify);

var _pgSql = require("pg-sql2");

var _pgSql2 = _interopRequireDefault(_pgSql);

var _debug = require("debug");

var _debug2 = _interopRequireDefault(_debug);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const debugSql = (0, _debug2.default)("graphile-build-pg:sql");

/*
 * Originally we tried this with a CTE, but:
 *
 * > The sub-statements in WITH are executed concurrently with each other and
 * > with the main query. Therefore, when using data-modifying statements in
 * > WITH, the order in which the specified updates actually happen is
 * > unpredictable. All the statements are executed with the same snapshot (see
 * > Chapter 13), so they cannot "see" one another's effects on the target
 * > tables. This alleviates the effects of the unpredictability of the actual
 * > order of row updates, and means that RETURNING data is the only way to
 * > communicate changes between different WITH sub-statements and the main
 * > query.
 *
 * -- https://www.postgresql.org/docs/9.6/static/queries-with.html
 *
 * This caused issues with computed columns that themselves went off and
 * performed selects - because the data within those selects used the old
 * snapshot and thus returned stale data.
 *
 * To solve this, we tried using temporary tables to ensure the mutation and
 * the select execute in different statments. This worked, but temporary tables
 * require elevated priviliges and thus don't work everywhere. We needed a more
 * generic solution.
 *
 * In the end we settled for sending the data we received from the mutations
 * straight back into the PostgreSQL server. It's a bit wasteful but it works.
 *
 * If you can come up with a better solution please open a pull request!
 */

exports.default = async function viaTemporaryTable(pgClient, sqlTypeIdentifier, sqlMutationQuery, sqlResultSourceAlias, sqlResultQuery, isPgClassLike = true) {
  async function performQuery(pgClient, sqlQuery) {
    // TODO: look into rowMode = 'array'
    const { text, values } = _pgSql2.default.compile(sqlQuery);
    if (debugSql.enabled) debugSql(text);
    return pgClient.query(text, values);
  }

  if (!sqlTypeIdentifier) {
    // It returns void, just perform the query!
    return await performQuery(pgClient, _pgSql2.default.query`
      with ${sqlResultSourceAlias} as (
        ${sqlMutationQuery}
      ) ${sqlResultQuery}`);
  } else {
    const result = await performQuery(pgClient, _pgSql2.default.query`
      with ${sqlResultSourceAlias} as (
        ${sqlMutationQuery}
      )
      select ${isPgClassLike ? sqlResultSourceAlias : _pgSql2.default.query`${sqlResultSourceAlias}.${sqlResultSourceAlias}`}::${sqlTypeIdentifier} from ${sqlResultSourceAlias}`);
    /*
     * This is a bit of a hack. Basically `pg` likes to send us JSON parsed,
     * but doesn't like it when we send JSON back through without stringifying.
     * We're not really mapping back to GraphQL here so we shouldn't use
     * pg2gql. So we're just going to JSON.stringify the value if we don't
     * think that `pg` will be happy with it.
     *
     * The optimium solution would be to have the `pg` module not parse the
     * data at all and just send it through to us as a string which we can then
     * post straight back to postgres. There may be options to enable this - if
     * you find them then a pull request would be welcome!
     */
    const { rows, fields } = result;
    const typeID = String(fields[0].dataTypeID);
    const jsonTypeId = "114";
    const jsonbTypeId = "3802";
    const mapValue = val => {
      if (val && (typeof val === "object" || typeID === jsonTypeId || typeID === jsonbTypeId)) {
        return (0, _stringify2.default)(val);
      } else {
        return val;
      }
    };
    const values = rows.map(row => mapValue(row[(0, _keys2.default)(row)[0]]));
    return await performQuery(pgClient, _pgSql2.default.query`
      with ${sqlResultSourceAlias} as (
        select ${isPgClassLike ? _pgSql2.default.query`(str::${sqlTypeIdentifier}).*` : _pgSql2.default.query`str::${sqlTypeIdentifier} as ${sqlResultSourceAlias}`}
        from unnest((${_pgSql2.default.value(values)})::text[]) str
      )
      ${sqlResultQuery}`);
  }
};
//# sourceMappingURL=viaTemporaryTable.js.map